/*
	Spectroscopic Toolkit version 1.40 by Pieter Suurmond, november 11, 2002.
	See headerfile 'ST_aiff.h'.
*/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include "ST_aiff.h"

/*--------------------------------------------------------------------------*/
/* Writes 2 byte short, MSB first. Works ok on both big and little endians. */
static short aiffWBshort (unsigned short u0, FILE* fp)
{
	if (EOF == putc(u0 >> 8, fp))
        return(-1);
	if (EOF == putc(u0 & 0xff, fp))
        return(-2);
    return(0);
}

/*------------------------------------------------*/
/* Writes 4 byte long, MSB first.                 */
static short aiffWBlong (unsigned long u0, FILE* fp)
{
    unsigned long u1 = u0 >> 8;
    unsigned long u2 = u1 >> 8;
	if (EOF == putc((u2 >> 8),  fp))
        return(-1);
	if (EOF == putc( u2 & 0xff, fp))
        return(-2);
	if (EOF == putc( u1 & 0xff, fp))
        return(-3);
	if (EOF == putc( u0 & 0xff, fp))
        return(-4);
    return(0);
}

/*-----------------------------------------------------------------------
    C O N V E R T   T O   I E E E   E X T E N D E D
    Machine-independent I/O routines for IEEE floating-point numbers.
    NaN's and infinities are converted to HUGE_VAL or HUGE, which
    happens to be infinity on IEEE machines.  Unfortunately, it is
    impossible to preserve NaN's in a machine-independent way.
    Infinities are, however, preserved on IEEE machines.
    These routines have been tested on the following machines:
        Apple Macintosh, MPW 3.1 C compiler
        Apple Macintosh, THINK C compiler
        Silicon Graphics IRIS, MIPS compiler
        Cray X/MP and Y/MP
        Digital Equipment VAX
    Implemented by Malcolm Slaney and Ken Turkowski.
    Malcolm Slaney contributions during 1988-1990 include big- and little-
    endian file I/O, conversion to and from Motorola's extended 80-bit
    floating-point format, and conversions to and from IEEE single-
    precision floating-point format.
    In 1991, Ken Turkowski implemented the conversions to and from
    IEEE double-precision format, added more precision to the extended
    conversions, and accommodated conversions involving +/- infinity,
    NaN's, and denormalized numbers.
-----------------------------------------------------------------------*/

#ifndef HUGE_VAL
# define HUGE_VAL HUGE
#endif /*HUGE_VAL*/

# define FloatToUnsigned(f) ((unsigned long)(((long)(f - 2147483648.0)) + 2147483647L) + 1)

static void ConvertToIeeeExtended(double num, char *bytes)
{
    int    sign;
    int expon;
    double fMant, fsMant;
    unsigned long hiMant, loMant;

    if (num < 0)
        {
        sign = 0x8000;
        num *= -1;
        }
    else
        {
        sign = 0;
        }
    if (num == 0)
        {
        expon = 0; hiMant = 0; loMant = 0;
        }
    else
        {
        fMant = frexp(num, &expon);
        if ((expon > 16384) || !(fMant < 1))
            {                                            /* Infinity or NaN */
            expon = sign|0x7FFF; hiMant = 0; loMant = 0; /* infinity */
            }
        else
            {                   /* Finite */
            expon += 16382;
            if (expon < 0)
                {               /* denormalized */
                fMant = ldexp(fMant, expon);
                expon = 0;
                }
            expon |= sign;
            fMant = ldexp(fMant, 32);          
            fsMant = floor(fMant); 
            hiMant = FloatToUnsigned(fsMant);
            fMant = ldexp(fMant - fsMant, 32); 
            fsMant = floor(fMant); 
            loMant = FloatToUnsigned(fsMant);
            }
        }
    bytes[0] = expon >> 8;
    bytes[1] = expon;
    bytes[2] = hiMant >> 24;
    bytes[3] = hiMant >> 16;
    bytes[4] = hiMant >> 8;
    bytes[5] = hiMant;
    bytes[6] = loMant >> 24;
    bytes[7] = loMant >> 16;
    bytes[8] = loMant >> 8;
    bytes[9] = loMant;
}

/*--------------------------------------------------------------------------------------------*/
static short aiffWriteHeader (AUDIOFILE* A)	/* Only used within this file, no need to call it */
{                                         	/* directly. Call aiffOpenForWriting() and        */
    short           b, e;                	/* and aiffCloseForWriting() instead.             */
    unsigned long   numBytes;				/* Returns -10, -11, -12, or -13 on failure.      */
    char            buf[10];

    /*---------------------------------------------------- Check header-info before writing:  */
    if ((A->channels < kAIFFWRITE_min_channels) || (A->channels > kAIFFWRITE_max_channels))
        return -10;										/* -10 = Wrong number of channels.    */
    b = A->bits >> 3;
    if ((A->bits & 7) || (b < 1) || (b > 4))
		return -11;										/* -11 = Wrong number of bits.        */
    if ((A->rate < kAIFFWRITE_min_samplerate) || (A->rate > kAIFFWRITE_max_samplerate))
        return -12;										/* -12 = Wrong samplerate.            */
    numBytes = A->frames * (unsigned long)b * A->channels;
	e = (EOF == fputs("FORM", A->fp));                  /* Write IFF header.                  */
	e |= aiffWBlong(8+18+8+12 + numBytes, A->fp);       /* COMM hdr + COMM chunk +            */
                                                        /* SSND hdr + SSND chunk + numBytes.  */
	e |= (EOF == fputs("AIFF", A->fp));                 /* File type                          */
    /*---------------------------------------------------- COMM chunk ------------------------*/
	e |= (EOF == fputs("COMM", A->fp));                 /* Describes encoding and #frames.    */
	e |= aiffWBlong ((long)18, A->fp);                  /* COMM chunk size.                   */
	e |= aiffWBshort(A->channels, A->fp);               /* Number of channels.                */
	e |= aiffWBlong (A->frames, A->fp);                 /* Number of sampleframes.            */
	e |= aiffWBshort(A->bits, A->fp);                   /* Sample width, in bits.             */
	ConvertToIeeeExtended((double)A->rate, buf);
    e |= (-10 + fwrite(buf, 1, 10, A->fp));             /* Samplerate in cycles per second.   */
	/*----------------------------------------------------- SSND chunk -----------------------*/
	e |= (EOF == fputs("SSND", A->fp));
	e |= aiffWBlong(8 + numBytes, A->fp);               /* Chunk size. */
	e |= aiffWBlong((long)0L, A->fp);                   /* Offset.     */
	e |= aiffWBlong((long)0L, A->fp);                   /* Block size. */
    if (e)
        e = -13;										/* -13 = Error (re)writing header.    */
    return e;
}


/*------------------------------------------------------------------------------------*/
short openAIFFWRITE (AUDIOFILE** 	A,			/* Handle receiving AUDIOFILE object. */
					 char*			filename,
					 long         	samplerate,
					 unsigned long	frames,    	/* Total number of frames (or NULL).  */
    				 unsigned short	channels,  	/* Number of interleaved channels.    */
    				 unsigned short	bits)      	/* Number of bits per channel.        */
{
    short   err;
												/* Check first 2 arguments.  */
	if (!A) return -1;							/* -1 = No handle.      	 */
	if (!filename) return -2;					/* -2 = No filename.    	 */
	if (!*filename) return -3;					/* -3 = Empty filename. 	 */
	*A = malloc(sizeof(AUDIOFILE));
	if (!*A) return -4;							/* -4 = Out of memory.              */
    (*A)->fp = fopen(filename, "wb");    		/* Try to create a new aiff file:   */
	if (!(*A)->fp)    							/* 'b' for platform-compatibility.  */
		{ free(*A); *A = NULL; return -5; }		/* -5 = Cannot create AIFF file.    */
	(*A)->rate     = samplerate;				/* Copy arguments, check later at   */
	(*A)->channels = channels;					/* aiffWriteHeader().               */
	(*A)->bits     = bits;
	if (frames)
		{
		(*A)->rewrite = 0;
		(*A)->frames  = frames;
		}
	else
		{
		(*A)->rewrite = 1;						/* If zero frames, needs rewrite!  */
		(*A)->frames  = kAIFFWRITE_max_frames;	/* Set header to max instead of 0, */
		}										/* so when one aborts without calling   */
	err = aiffWriteHeader(*A);					/* closeAIFFWRITE() any content is visible. */ 
	(*A)->frames = frames;						/* Set back to 0 again if it was 0. */
    if (err)									/* Error -10, -11, -12, or -13.   */
        { fclose((*A)->fp); free(*A); *A = NULL; }
    return(err);
}

/*--------------------------------------------------------------------------------------*/
short dumpAIFFWRITE (AUDIOFILE* A, void* audio, unsigned long frames)
{
	size_t	written;        	/* Use type long or unsigned long if size_t is unknown. */

	if (!(A && audio && frames))
		return -1;			     	/* Num bytes per frame. */
	written = fwrite(audio, (size_t)(A->bits >> 3) * A->channels, (size_t)frames, A->fp);
	A->frames += written;          			/* Advance our global frame-counter.   */
   	if (A->frames > kAIFFWRITE_max_frames)	/* Extra care for your disk drives.    */
    	return -2;
	if (written != frames)
    	return -3;
	return 0;
}

/*----------------------------------------------------------------------*/
void printAIFFWRITE (AUDIOFILE* A, FILE* to)
{
    double          secs;
    const char*     line = "------------------------------------------\n";
	if (A && to)
		{
    	fprintf(to, line);
    	fprintf(to, "        channels: %d\n",     A->channels);
    	fprintf(to, "          frames: %ld\n",    A->frames);
    	fprintf(to, "            bits: %d\n",     A->bits);
    	fprintf(to, "      samplerate: %ld Hz\n", A->rate);
    	/*--------------------------------------- DERIVED INFO: ----------*/
    	fprintf(to, " bytes per frame: %d\n", A->channels * (A->bits >> 3));    
    	fprintf(to, "        filesize: %ld bytes\n",
                	8+18+8+12 + (A->frames * (A->bits>>3) * A->channels));
    	fprintf(to, "        duration: ");
    	secs = (double)A->frames / (double)A->rate;
    	if (secs >= 3600.0)	fprintf(to, "%.2f hours\n", secs/3600.0);
    	if (secs >= 60.0)   fprintf(to, "%.2f minutes\n", secs/60.0);
    	else                fprintf(to, "%.2f seconds\n", secs);
    	fprintf(to, line);
		}
}

/*------------------------------------------------------------------------------------*/
short closeAIFFWRITE (AUDIOFILE** A)
{
    short 	err = 0;

	if (!A) return -1;						/* -1 = No handle.      */
	if (!*A) return -2;						/* -2 = Already closed. */
    if ((*A)->rewrite)
        {
        if (fseek((*A)->fp, 0L, 0))
            err = -3;						/* -3 = Can't rewind to rewrite. */
      	else
        	err = aiffWriteHeader(*A);		/* -10, -11, -12, or -13. */
        }
    if (fclose((*A)->fp))
        err += -100;						/* Closing AIFF file failed. */
	free(*A);								/* Sure it's there. */
	*A = (AUDIOFILE*)NULL;
    return(err);
}
