/*
	Spectroscopic Toolkit version 1.60 by Pieter Suurmond, december 1, 2002.
    
    Internal structure definitions of WAVELET and ENGINE. Forward declarations are in header
    ST_wavelets.h, which must be included before this file can be included.

	The private definitions here are included in internal sourcefiles ST_wavelets.c 
    and in ST_score.c ONLY, this file should NOT be visible to the user.
*/

/*---------------------------------------- LINKED LIST STRUCTURE FOR WAVELETS: ---------------*/
struct WAVELET                          /* Basic idea is that wavelets are sorted on start-   */
{                                       /* time so playing to AIFF can be done efficiently.   */
    WAVELET     nxt;                    /* Pointer to next WAVELET object or NULL (the end).  */
    UNSIGNED32  startFrame;             /* Absolute sampleframe time calculated from sample-  */
    UNSIGNED32  numFrames;              /* rate, TOP and DUR by the addWAVELET() routine.     */
    SIGNED32    amplitudeLeft;          /* 24 bit linear from -8388608 to +8388608 incl.,     */
    SIGNED32    amplitudeRight;         /* calculated from PAN and AMP by addWAVELET().       */
	SIGNED16	envelopeShape;		    /* Choosen envelope shape. One of the constants       */
									    /* kENV_Gaussian, kENV_ExpDecay, or kENV_ExpRise.     */
    UNSIGNED64  envelopeTime;           /* 64 bit fixed-point: 18 integer + 46 fractional.    */
    UNSIGNED64  envelopeStep;           /* Calculated from DUR, ENV and samplerate.           */
    UNSIGNED64  oscillatorTime;         /* 64 bit fixed-point: 18 integer + 46 fractional.    */
                                        /* Initialiased by addWAVELET(), in sync or to zero.  */
                                        /* Unsigned to prevent sign-extension at shift-right. */
    UNSIGNED64  oscillatorStep;         /* For extreme frequency resolution: Kurucz's files   */
};                                      /* contain 10 wavelength-digits so we need 33.22      */
                                        /* (thus 34) bits. Calculated from FRQ and sample-    */
                                        /* rate by addWAVELET().                              */

/*------------------------------------ DSP ENGINE: -------------------------------------------*/
struct ENGINE                       /* An engine is larger than a wavelet.                    */
{
    long            samplerate;         			/* Samplerate related constants, initia-  */
    double          oneOverSamplerate64;			/* lised by newENGINE().                  */
    double          fourOverSamplerate;
	UNSIGNED32	    framesPerBlock;					/* Amount of frames per block at rende-   */
												    /* ring and block-transfer to AIFF file.  */
    SIGNED32        *sineTable;       				/* Audio and subaudio wavetables, initia- */
    SIGNED16        *envTables[kNumEnvelopeShapes];	/* lised by newENGINE(), read out by      */
												    /* playWAVELETS().                        */

    struct WAVELET  rootWAVELET;            /* Anchor for single linked list. newENGINE()     */
										    /* gives an empty list, addWAVELET() adds wave-   */
										    /* lets, playWAVELETS() removes them again.       */

                                    	    /* Statistics kept up to date by addWAVELET():    */
    UNSIGNED32      WSTAT_number;       	/* Number of wavelets in list, just for fun.      */
    UNSIGNED32      WSTAT_lastFrame;    	/* Not necessarily belonging to LAST wavelet      */
    UNSIGNED64      WSTAT_loOscillStep, 	/* in the list (ENDtime)! First frame, on the     */
                    WSTAT_hiOscillStep; 	/* contrary, equals starttime first wavelet.      */
};

/*--------------------------------- SOME CONSTANTS: ------------------------------------------*/
#define     kTwoPi                  ((double)6.2831853071795862)          /* 8.0 * atan(1.0). */
#define     k2raise64               ((double)18446744073709552000.0)      /* 2^64. */

/*--------------------------------- 256 K samples for audio sinewave. With 17 bit signed data
                                    steepest slope is 0.7854 bit. Requires 18 addressing bits
                                    (2^18=262144). 64 bits are used for frequency so 46 bits
									remain for the fractional part. Constant kSineTableSize
									must be a power of 2.           
									Minimum = -32768: 11111111 11111111 10000000 00000000.
									Maximum = +32768: 00000000 00000000 10000000 00000000.    */
#define     kSineTableSize          (262144)              
#define     kSineTableFractBits     (46)
#define     kSineTableHalfOffset	((UNSIGNED64)1 << (kSineTableFractBits - 1))

/*--------------------------------- 256 K samples for envelopes. With 16 bit gaussian content
                                    steepest slope is 0.75814 bit. Data always positive, data-
                                    type is however signed, stored negatively to fit top-
                                    value -32768 within 16 bits.  All envelope tables require 
                                    18 bit addressing (like the sine-table above). In case of
									exponential envelope, steepest slope is 1.3 bit.
									Maximum =      0:  00000000 00000000.
									Minimum = -32768:  10000000 00000000.                     */
#define     kEnvTableSize           (262144)
#define     kEnvTableFractBits      (46)
