/*
	Spectroscopic Toolkit version 1.60 by Pieter Suurmond, november 26, 2002.
	See headerfile ST_wavelets.h.
*/

#include <stdio.h>              /* Standard headers needed here. */
#include <stdlib.h>
#include <math.h>

#include "ST_aiff.h"            /* Need to access filepointer, samplerate and samplesize.  */
#include "ST_wavelets.h"
#include "ST_engine.h"          /* Private definition of ENGINE struct. */

/*-----------------------------------------------------------------------------------------*/
short addWAVELET(ENGINE E,    FILE*  MSG,								/* See headerfile. */
				 double FRQ,  double AMP,  double PAN,  double TOP,  double DUR,
				 short	ENV,  short  SYN)
{
    WAVELET     fresh, search, previous;
    UNSIGNED32  frame, frame2;
	UNSIGNED64	phase;
    double      calculate, updown, realDUR;
	
	if (!E) return 1;
    if (FRQ >= 0.5*(double)E->samplerate)           	/* Must be below Nyquist.        */
        {
        if (MSG) fprintf(MSG, "FRQ (%.6f Hz) not below Nyquist (%.6f Hz)!\n",
                         FRQ, 0.5*(double)E->samplerate); return 2;
        }
    if (FRQ < 16.0)                         	/* Not below 16 Hz and not negative!     */
        {                 		/* (There are DSP-reasons why FREQ may not be negative). */
        if (MSG) fprintf(MSG, "FRQ too low (%.6f Hz)!\n", FRQ); return 3;
        }
    calculate = fabs(AMP);   	/* Contrary to FRQ, AMP may be negative,       */
    if (calculate > 1.0)     	/* so if you DO want to invert phase, use AMP. */
        {
        if (MSG) fprintf(MSG, "AMP too high (%.6f)!\n", AMP); return 4;
        }
    if (calculate < (double)(1.0 / 8388608.0))   /* Store at least 23 bits of magnitude. */
        {
        if (MSG) fprintf(MSG, "AMP too low (%.6f)!\n", AMP); return 5;
        }
    if (fabs(PAN) > 1.0)                            /* Between -1 and +1. */
        {
        if (MSG) fprintf(MSG, "PAN out of range (%.4f)!\n", PAN); return 6;
        }      
    if (DUR < E->fourOverSamplerate)                /* Minimal 4 samples. */
        {
        if (MSG) fprintf(MSG, "DUR too low (%.6f ms)\n", 1000.0 * DUR); return 7;
        }
    if (DUR > 3600.0)                               /* Max 1 hour for convenience.   */
        {                                           /* So ACTUAL max. dur=5 hours.   */
        if (MSG) fprintf(MSG, "DUR too high (%.4f s)\n", DUR); return 8;
        }
	if ((ENV < 0) || (ENV >= kNumEnvelopeShapes))	/* 0,1 or 2. */
		{
        if (MSG) fprintf(MSG, "ENV unknown (%d)!\n", ENV); return 9;
        }
    calculate = TOP;                  				/* Check starttime. */
	if (ENV == kENV_Gaussian)
		{
		realDUR = 5.0 * DUR;
		calculate -= (2.5 * DUR);
		}
	else /* THUS exponential: */
		{
		if (SYN) /* Illegal ENV-SYN-combination? */
			{
			if (MSG) fprintf(MSG, "Exponential wavelets must have SYN=0!\n");
			return 10;
			}
		realDUR = 15.0 * DUR;
		if (ENV == kENV_ExpRise)
			calculate -= realDUR;
		}
    if (calculate < 0.0)                            /* Only positive starttimes.     */
        {
        if (MSG) fprintf(MSG, "TOP too early (starttime = %.4f seconds)!\n", calculate);
      	return 11;
        }
    if (calculate > 86400.0)                                /* Max 24 hours. */
        {
        if (MSG) fprintf(MSG, "TOP too late (starttime = %.2f hours)!\n", calculate / 3600.0);
     	return 12;
        }
    if (!(fresh = malloc(sizeof(struct WAVELET))))
        {
        if (MSG) fprintf(MSG, "Not enough memory to allocate wavelet!\n"); return 13;
        }
    /*-------------------------------------------------------- OK, so far no errors: --------*/
    if (MSG)
        fprintf(MSG, "Allocating wavelet FRQ=%12.6f AMP=%7.4f PAN=%4.1f TOP=%.4f DUR=%.4f.\n",
                FRQ, AMP, PAN, TOP, DUR);  /* Notice real duration is 5 or 15 times as long! */
    /*-------------------------------------------------------------------- Envelope timing:  */
	fresh->envelopeShape = (SIGNED16)ENV;
    fresh->startFrame    = (UNSIGNED32)(0.5 + ((double)E->samplerate * calculate));
	fresh->numFrames     = (UNSIGNED32)(0.5 + ((double)E->samplerate * realDUR));
	fresh->envelopeStep  = (UNSIGNED64)(k2raise64 / (double)fresh->numFrames); /* DOWN! */

	/* For the rising exponential, we must be sure that the end of the envelope table   */
	/* is exactly reached! Futhermore, the audio-oscillator needs to be aligned.        */
	/* Contrary to the audio table, we do NOT want the envelope table to wrap while     */
	/* reaching its' final sample. Sometimes (when ENV == kENV_ExpRise) we want the     */
	/* envelope phase to reach exactly the last sample.	*/

	phase = (UNSIGNED64)(fresh->numFrames - 1) * fresh->envelopeStep;
	if ((phase <= (UNSIGNED64)0x8000000000000000) && MSG)
		fprintf(MSG, "WARNING: envelope table wrap!\n");
	if (ENV == kENV_ExpRise)								/* No compensation yet for gaussians! */
		{
		fresh->envelopeTime = 0xFFFFFFFFFFFFFFFF - phase;	/* Maximal compensation for reverse envelopes. */
		frame  = fresh->startFrame + fresh->numFrames - 1;	/* Last frame actually calculated and written. */
		frame2 = (long)(0.5+(TOP * (double)E->samplerate));	/* Specified frame (rounded to nearest frame). */
		if (frame < frame2)
			{
			fresh->startFrame++;							/* To let TOP equal last calculated frame.     */
			frame++;										/* Just for the next check. */
#if 0		/* Print starttime correction for debugging: */
			if (MSG) fprintf(MSG, "startFrame++ compensation.\n");
#endif
			}
		if ((frame != frame2) && MSG)
			fprintf(MSG, "WARNING: specified TOP = %ld but last frame = %ld!\n", frame2, frame);
#if 0 	/* Print envelope phase compensations for debugging: */
		if (MSG)
			{
			fprintf(MSG, "OFFSET ->envelopeTime INT = %lld\n",   (fresh->envelopeTime >> kEnvTableFractBits));
			fprintf(MSG, "       ->envelopeTime FRC = %lld\n\n", (fresh->envelopeTime &  0x00003FFFFFFFFFFF));
			}
#endif
		}
	else 									  			/* All other envelopes start at 0.0. */
		fresh->envelopeTime = (UNSIGNED64)0;		    /* (not 0.5 ! to prevent wrap)       */
    /*-------------------------------------------------------------------- Audio frequency:  */
    fresh->oscillatorStep = (UNSIGNED64)(0.5 + (FRQ * E->oneOverSamplerate64));
	fresh->oscillatorTime = kSineTableHalfOffset;      /* Always positive, already add half. */
	if (SYN) /* All oscills perfectly in sync at time 0, 64b-multiplic. may safely overflow. */
		fresh->oscillatorTime += (UNSIGNED64)fresh->startFrame * fresh->oscillatorStep;
	else if (ENV == kENV_ExpRise) 	   /* Then SYN is never set. / Last phase actually used. */
		fresh->oscillatorTime -= ((UNSIGNED64)(fresh->numFrames - 1) * fresh->oscillatorStep);
    /*-------------------------------------------------------------------- Panning -1 to 1:  */
    AMP *= 8388608.0;    /* 2^23 = 00000000 10000000 00000000 00000000	   24 bits + sign.   */
    if (AMP >= 0.0)      /* Round correctly . */
        updown = +0.5;
    else
        updown = -0.5;
    fresh->amplitudeLeft  = (SIGNED32)(updown + (AMP * sqrt(0.5 * (1.0 - PAN))));
    fresh->amplitudeRight = (SIGNED32)(updown + (AMP * sqrt(0.5 * (1.0 + PAN))));
    /*-------------------------------------------------------------------- Keep up stats:    */
    E->WSTAT_number++;  /* Don't depend on this number... (yet,.. we've never needed it).    */
    frame = fresh->startFrame + fresh->numFrames;
    if (frame > E->WSTAT_lastFrame)                         /* (init @ 0).   */
        E->WSTAT_lastFrame = frame;                         /* (downrounded) */
    if (fresh->oscillatorStep < E->WSTAT_loOscillStep)      /* (init @ 0xFFFFFFFFFFFFFFFF).  */
        E->WSTAT_loOscillStep = fresh->oscillatorStep;
    if (fresh->oscillatorStep > E->WSTAT_hiOscillStep)      /* (init @ 0).   */
        E->WSTAT_hiOscillStep = fresh->oscillatorStep;
    /*-------------------------------------------------------------------- List management:  */
    search = &E->rootWAVELET;           /* Pointer to wavelet with dummy-content.            */
    previous = search;
    while ((search = search->nxt))      /* Maintain ordering on starttime so we don't need   */
        {                               /* to sort the linked list of wavelets afterwards.   */
        if (fresh->startFrame < search->startFrame)
            break;
        previous = search;              /* Forget previous, remember this one. */
        }
    previous->nxt = fresh;
    fresh->nxt = search;                /* Successive wavelet or NULL. */
    return 0;                           /* return 0 for ok.  */
}

/*-----------------------------------------------------------------------------------------*/
short playWAVELETS(ENGINE E, char* filename, short bits, FILE* MSG, FILE* releaseMSG)
{					
    short      	error, error2;
    long      	bytes;          	/* Number of bytes of the big buffer (dynamic alloc.). */
    short     	bytesPerChannel;	/* Number of bytes of the small buffer (= bits / 8).   */
    SIGNED8  	outputBytes[4]; 	/* The small static buffer itself, maximal 32 bits.    */
    short    	b, s, shiftR;
    SIGNED8   	*interleavedRewritePtr;             	/* Quick 'n dirty == beautiful?    */
    SIGNED64 	*interleavedAccu, *interleavedPtr;  	/* Memory to mix and buffer audio. */
    SIGNED32  	a;                                  	/* Enveloped but unpanned signal.  */
	SIGNED16	*e;
    UNSIGNED64	oTime, eTime;
    UNSIGNED32	n, skip, samplesPerBlock;
    SIGNED64 	rounded, half,fractMask, uFlo,oFlo; /* call, otherwise overruled by    */
    WAVELET    	begin, previous;                    /* supplied gOUTFILE.frames-value. */
	AUDIOFILE	*A;
																	   /* (2 means stereo) */
	error = openAIFFWRITE(&A, filename, E->samplerate, (unsigned long)0, 2, bits);
	if (error)
		{ if (releaseMSG) fprintf(releaseMSG, "openAIFFWRITE() = %d!\n", error); return 1; }
    if (releaseMSG)
		fprintf(releaseMSG, "\nplayWAVELETS() dumping blocks of %ld frames to file '%s'.\n",
					        (long)E->framesPerBlock, filename);
    /*-------------------------------------------------------- PREPARE SOME SCALING-STUFF: */
    bytesPerChannel = bits >> 3;                         	/* Bytes = bits / 8.           */                                                   /* Determine how many times    */
    shiftR = 54 - bits;                                  	/* Determine how many times    */
    fractMask = ((SIGNED64)0x003FFFFFFFFFFFFF) >> bits;  	/* output has to rightshifted. */
	half =      ((SIGNED64)0x0020000000000000) >> bits;  	/* Determine where the roun-   */
    /*  8 bit:    0x00qq3FFFFFFFFFFF.                          ding-bit (0.5) should sit.  */
	/*            0x00qq200000000000.                                                      */
    /*              18 +1       +45   >>46 for 8 bit output.                          	   */
    /* 16 bit:    0x00qqqq2000000000.                   	   'half' is added or subtrac- */
	/*              26   +1     +37   >>38 for 16 bit output.   ted after all is added up. */
    /* 24 bit:    0x00qqqqqq20000000.                   	                               */
    /*              34     +1   +29   >>30 for 24 bit output.                              */
    /* 32 bit:    0x00qqqqqqqq200000.                                                      */
    /*              42       +1 +21   >>22 for 32 bit output.                              */
    s = 32 - bits;
    oFlo = ((SIGNED64)0x000000007FFFFFFF) >> s;     /* Adjust over- and underflow accor-   */
    uFlo = ((SIGNED64)0xFFFFFFFF80000000) >> s;     /* ding to the AIFF output samplesize. */
	/*---------------------------------- ALLOCATE 64 BIT INTERLEAVED ACCUMULATOR BUFFER: --*/
	samplesPerBlock = E->framesPerBlock * A->channels;
    bytes = samplesPerBlock * sizeof(SIGNED64);
    if (!(interleavedAccu = malloc(bytes)))                 			 /* *2 for stereo. */
        {
        if (MSG) fprintf(MSG, "Not enough memory for 64 bit buffer (%ld bytes)!\n", bytes);
        return 2;                           /* Don't exit without freeing mem from now on! */
        }
    while (E->rootWAVELET.nxt && (!error))  /* As long as linked list contains wavelets.   */
        {
        interleavedPtr = interleavedAccu;   /* Cleanup 64-bit stereo interleaved buffer.   */
        for (n = samplesPerBlock; n--;)		/* +0.5 disregarding sign! See the correction  */
            *interleavedPtr++ = half;		/* +0.5 disregarding sign(!)                   */
        /*---------------------------- Add up all wavelets within this section of time: ---*/
        begin = &E->rootWAVELET;            /* Start at the beginning of the list with our */
        previous = begin;                   /* our dummy-wavelet (only pointer to next     */
        while ((begin = begin->nxt))        /* is relevant).                               */
            {
            if (begin->startFrame >= (A->frames + E->framesPerBlock))
                goto endLL;                 	/* Starttimes in ascending order, we're    */
                                            	/* done now with this section of time.     */           
            n = E->framesPerBlock;       		/* Default is the whole accumulator.       */
            interleavedPtr = interleavedAccu;
			if (begin->startFrame > A->frames)	/* Wavelet not already started in previous */
				{								/* block (>= should also work correctly).  */
                skip = begin->startFrame - A->frames;		/* Amount of frames preceding  */
                interleavedPtr += (skip * A->channels);  	/* start. Twice for stereo.    */
                n -= skip;
                }   /* addWAVELET() already initialised ->oscillatorTime to be in sync.    */
           	oTime = begin->oscillatorTime;   		/* Load contents in cpu (before loop). */
            eTime = begin->envelopeTime;            /* oTime & eTime maybe register-vars?  */
			e = E->envTables[begin->envelopeShape];	/* Never out of range.                 */
            /*-------------------------------- Add this wavelet to accu-buffer: -----------*/
            while (n--)                     /* Initially, n never 0, however, would a      */
                {                           /* do-while with pre-dec be more efficient?    */       
                if (!begin->numFrames--)    /* Might be initially 0, is however limited.   */
                    {           /* Cut-out this wavelet, we're completely done with it now.*/
                    if (releaseMSG) fprintf(releaseMSG, "\
Releasing wavelet: start=%.2f s, dur=%.2f s, f=%.2f Hz.\n",                 /* Link previ- */
                     (double)begin->startFrame / (double)E->samplerate,     /* ous to next */
                     E->oneOverSamplerate64 / (double)begin->envelopeStep,  /* (or NULL!). */
                     (double)begin->oscillatorStep / E->oneOverSamplerate64);
                    previous->nxt = begin->nxt; /* Thus skipping this one. Immediately re- */
                    free(begin);                /* lease memory so list destroys 'itself'. */
                    begin = previous;           /* Behave as if we've just finished the    */
                    goto endWL;                 /* preceding one, but do NOT try to write  */
                    }	                        /* times back into memory (they're gone!). */
                /*-------------------------------- Calculate enveloped mono signal: -------*/
                /* 17*16 bit multiply. Upper 18 bits of oTime and eTime are used to index  */
                /* tables. No interpolation necessary at the table-lookups. Unsigned indi- */
                /* ces so we do not need to mask-off sign bits (no sign-extension).        */
				a = E->sineTable[oTime >> kSineTableFractBits]
                    * (SIGNED32)(e[eTime >> kEnvTableFractBits]);
                /*-------------------------------- Enforce 64 * 64 bit multiplication: ----*/
                /*                                 by (SIGNED64)-cast.      16*15*23 gives */
               	(*interleavedPtr++) -= (SIGNED64)a * begin->amplitudeLeft;  /*    54 bits. */
            	(*interleavedPtr++) -= (SIGNED64)a * begin->amplitudeRight; /* See shiftR! */
                oTime += begin->oscillatorStep;     /* Advance fractional table-pointers,  */
                eTime += begin->envelopeStep;       /* hopefully in cpu-reg (64 bit add).  */
                }
            begin->oscillatorTime = oTime;  /* Move both audio-time and envelope-time back */
            begin->envelopeTime   = eTime;  /* into memory again for the next block.       */
endWL:      previous = begin;               /* Forget the previous one, remember this one. */
            }
        /*------------------------------------ Round to X bits, rewrite buffer in-place: --*/
endLL:  interleavedPtr = interleavedAccu;               	/* Pointers initially same.    */
        interleavedRewritePtr = (SIGNED8*)interleavedAccu;
        for (n = 0; n < samplesPerBlock; n++)  				/* Interleaved buffer.         */
            {                                           	/* 'half' was already added.   */
            rounded = *interleavedPtr++;					/* So subtract 0.5 again for   */
			if ((rounded <= 0) &&							/* this special case. 		   */
				((rounded & fractMask) == (SIGNED64)0))		/* ST.wavelets.h explains why. */
				rounded -= half;							/* (We could sub 1.0 instead.) */
			rounded >>= shiftR;								/* Down to 8, 16, 24 or 32 bit.*/
            if (((rounded > oFlo) || (rounded < uFlo)) &&  	/* In case of over/underflow   */
				(!error))								/* rendering will be aborted at    */
                {                                       /* the next transfer-calc-cycle.   */
				if (MSG) fprintf(MSG, "Overflow (%lld) in channel %d at %.3f seconds!\n", 
						 rounded, (int)(n % A->channels), 
					     (double)(A->frames + (n / A->channels)) / (double)E->samplerate);
            	error = 3;                          	/* Remember only FIRST overflow.   */
                }
            /*---------------------------------------- WRITE BYTE FOR BYTE BACK IN BUFFER: */
            outputBytes[0] = (SIGNED8)rounded;      /* Take LSByte (& 0xFF not necessary). */
            b = bytesPerChannel;                    /* bytesPerChannel >= 1 ( = bits / 8). */
            s = 1;                                  /* Pre-decremented in do-loop below.   */
            while (--b)                             /* Copy LSB...MSB into small buffer.   */
                {
                rounded >>= 8;
                outputBytes[s++] = (SIGNED8)rounded;/* Take LSByte (& 0xFF not necessary). */
                }
            do  {                                   /* Write MSB...LSB back to large buff. */
                *interleavedRewritePtr++ = outputBytes[--s];
                } while (s);                        /* Overwriting part of previous data.  */
            }  /* This disk-transfer should work correctly on both BIG and LITTLE ENDIANS. */
        /*-------------------------------------------- Block-transfer to disk: ------------*/
		error2 = dumpAIFFWRITE(A, (void*)interleavedAccu, E->framesPerBlock);
		if (error2)            /* Framesize calculated by dumpAIFFWRITE(). */
			{				   /* Field A->frames is incremented.          */
			if (MSG) fprintf(MSG, "dumpAIFFWRITE() = %d!\n", error2);
			if (!error) error = 4;
			}
        }
    free(interleavedAccu);                	/* Release memory allocated above. */
	printAIFFWRITE (A, MSG);				/* Both args may be NULL. */
	if (releaseMSG != MSG)					/* Not twice to the same stream. */
		printAIFFWRITE (A, releaseMSG);
	error2 = closeAIFFWRITE(&A);
	if (error2)
		{
		if (MSG) fprintf(MSG, "closeAIFFWRITE() = %d!\n", error2);
		if (!error) error = 5;
		}
	if (releaseMSG)	/* More descriptive message already reported to MSG. */
		{
		if (error) fprintf(releaseMSG, "playWAVELETS() = %d!\n", error);
    	else fprintf(releaseMSG, "Ok, playWAVELETS() closed AIFF file '%s'.\n", filename);
		}
    return (error);
}

/*--------------------------------------------------------------------------------*/
void releaseWAVELETS(WAVELET begin)     /* Destructs all wavelets in linked list. */
{                                       /* Argument MAY be NULL.                  */
    WAVELET following;
    while(begin)
        {
        following = begin->nxt;
        free(begin);
        begin = following;
        }
}

/*--------------------------------------------------------------------------------------*/
static SIGNED32* newSineTable(FILE* MSG)				/* Only used within this file.  */
{
    SIGNED32        *C, *sinoid;                        /* malloc will set C initially. */
    double          wave, timestep;                     /* Cosine-table 32 bit content. */
    UNSIGNED32      t;                                  /* 32 bit to index tables.      */
    
    C = malloc(kSineTableSize * sizeof(SIGNED32));
    if (C)
        {
        sinoid = C;
        timestep = kTwoPi / (double)kSineTableSize;
        for (t=0; t<kSineTableSize; t++)
            {
            wave = 32768.0 * sin((double)t * timestep); /* Top 17 signed bits needed.   */
            if (wave >= 0.0)
                wave += 0.5;
            else
                wave -= 0.5;
            *sinoid++ = (SIGNED32)wave;             	/* 32 bits needed. */
            }
        if (MSG) fprintf(MSG, "Created sine-table (audio 32 bit, %ld samples)\n",
						 (long)kSineTableSize);
        }
    else if (MSG) fprintf(MSG, "Not enough memory for sine-table (audio)!\n");
    return(C);
}

/*----------------------------------------------------------------------------------------*/
static SIGNED16* newGaussTable(FILE* MSG)				   /* Only used within this file. */
{
    SIGNED16        *G,                             /* malloc inits G */
                    env, *content, *contentMirror;
    double          timestep, gt;
    UNSIGNED32      t;                              /* 32 bit to index tables. */
    /*-------------------------------------------- GUASSIAN ENVELOPE ---------------------*/
    /*  y = e ^ -(x^2)       start at x = -sqrt(25/2) --> y = 3.7267*10^-6                */
    /*                         end at x = +sqrt(25/2)    (about 0.12 at 15 bit unsigned.) */
    /*  Inflection points at          x = +sqrt(1/2)  equals 0.4 of the total tablesize.  */
    /*                                x = -sqrt(1/2)         0.6 tablesize.               */
	/* Amplitude at TOP (where x=0) is exactly 1.0. At the inflection points, amlitude is */
	/* e ^ -0.5 which is approximately 0.6065. We call DUR the effective duration, which  */
	/* is the time between the inflection points. The actually calculated duration is     */
	/* 5 times as long. One of the -6 dB point lies at:								      */
	/*		0.5 = e ^ (-xh^2)								          					  */
	/*		ln(0.5) = -xh^2								          					      */
	/*		-ln(0.5) = xh^2								          					      */
	/*		xh = sqrt(-ln(0.5)) = 0.832554611											  */
    /*------------------------------------------------------------------------------------*/
    G = malloc(kEnvTableSize * sizeof(SIGNED16));
    if (G)
        {
        timestep = sqrt(50.0) / (double)kEnvTableSize;
        content = G + (kEnvTableSize >> 1);   				/* Start of upper half. */
        contentMirror = content;    
        for (t=0; t<(kEnvTableSize >> 1); t++)
            {
            gt = (double)t * timestep;
            gt = -1.0 * gt * gt;
            env = (SIGNED16)((-32768.0 * exp(gt)) - 0.5);   /* Stored negative!!! */
            *content++ = env;                               /* Symmetrical around middle  */
            *contentMirror-- = env;                         /* Index 0 IS used so there   */
            }                                               /* will be time-err of 1 samp */
        *contentMirror = 0;
        if (MSG) fprintf(MSG, "Created symmetrical gauss-table (subaudio 16 bit, %ld samples)\n",
						 (long)kEnvTableSize);
        }
    else if (MSG) fprintf(MSG, "Not enough memory for gauss-table (envelope)\n");
    return(G);
}

/*----------------------------------------------------------------------------------------*/
static SIGNED16* newExpTable(short rise, FILE* MSG) /* Only used within this file. Decay- */
{													/* ing if argument rise is zero,      */
    SIGNED16        *G, *content;					/* rising if rise is nonzero.         */
    double          timestep;
    UNSIGNED32      t, tt;                              	   /* 32 bit to index tables. */
	const char*		riseTxt[2] = { "rising", "decaying" };
    /*-------------------------------------------- EXPONENTIAL DECAY ---------------------*/
    /*  y = e ^ -x      start at x = 0       --> y = 1                                    */
    /*                    end at x = ln(0.5) --> y = 0.5                                  */
	/* Amplitude at TOP (where x=0) is exactly 1.0.                                       */
	/* We call DUR the effective duration, which is the time between the top (1.0) and    */
	/* the -6 dB point (0.5). The actually calculated duration is 15 times as long.       */
    /*------------------------------------------------------------------------------------*/
    G = malloc(kEnvTableSize * sizeof(SIGNED16));
    if (G)
        {
		content = G;
        timestep = 15.0 * log(0.5) / (double)kEnvTableSize;	   /* 'log()' is natural log! */
        for (t=0; t<kEnvTableSize; t++)
            {								    					/* Stored negative!!! */
			if (rise) tt = kEnvTableSize - 1 - t; else tt = t;
            *content++ = (SIGNED16)((-32768.0 * exp((double)tt * timestep)) - 0.5);
            }
        if (MSG) fprintf(MSG, "Created %s exp-table (subaudio 16 bit, %ld samples)\n",
						      riseTxt[rise], (long)kEnvTableSize);
        }
    else if (MSG) fprintf(MSG, "Not enough memory for exp-table (envelope)\n");
    return(G);
}

/*----------------------------------------------------------------------------------------*/
ENGINE newENGINE(long sr, long blocksize, FILE* msg)    /* See headerfile ST.wavelets.h.  */
{
    ENGINE  E = NULL;

    if ((sr < 1L) || (sr > 144000L))					/* Be tolerant :-).               */
        { if (msg) fprintf(msg, "Invalid samplerate (%ld Hz) for wavelet-engine!\n", sr); }
	else if ((blocksize < 1L) || (blocksize > 1048576L))/* Not more than 1 million frames.*/
		{ if (msg) fprintf(msg, "Invalid blocksize (%ld sampleframes)!\n", blocksize); }
    /*------------------------------------ First a small datasize-check: -----------------*/
    else if ((sizeof(SIGNED16)<2) || (sizeof(UNSIGNED32)<4) || (sizeof(UNSIGNED64)<8) || 
             (sizeof(float)<4)    || (sizeof(double)<8)     || (sizeof(char)!=1))
        { if (msg) fprintf(msg, "Datatypes incompatible with platform / compiler!\n"); }
        /* Which should better be tested @ compile time instead of runtime. */
    /*------------------------------------ Allocate & init DSP-engine itself: ------------*/
    else
        {
        E = malloc(sizeof(struct ENGINE));
        if (E)
            {
            E->samplerate = sr;                             /* Samplerate-related constants.  */
            E->oneOverSamplerate64 = k2raise64 / (double)sr;/* Used by addWAVELET(), play-    */
            E->fourOverSamplerate =  4.0 / (double)sr;      /* WAVELETS(), scoreWAVELETS().   */
            E->framesPerBlock = blocksize;
		    E->rootWAVELET.nxt = (WAVELET)NULL;             /* Start with no wavelets.        */
            E->WSTAT_number = 0L;                           /* Incremented by addWAVELET().   */
            E->WSTAT_lastFrame = 0L;                        /* Updated by addWAVELET().       */
            E->WSTAT_loOscillStep = 0xFFFFFFFFFFFFFFFF;
            E->WSTAT_hiOscillStep = 0x0000000000000000;
            /*------------------------------------------------ ALLOCATE AND FILL WAVETABLES: -*/
		    E->sineTable                = newSineTable(msg);
		    E->envTables[kENV_Gaussian] = newGaussTable(msg);
		    E->envTables[kENV_ExpDecay] = newExpTable(0, msg);
		    E->envTables[kENV_ExpRise]  = newExpTable(1, msg);
		    if (!(E->sineTable &&                           /* All 4 must be there. */
			      E->envTables[kENV_Gaussian] &&
			      E->envTables[kENV_ExpDecay] &&
			      E->envTables[kENV_ExpRise]))
                {
			    releaseENGINE(&E);							/* Also sets E to NULL. */
                if (msg) fprintf(msg, "Not enough memory to allocate wavetables!\n");
                }
            }
        else if (msg) fprintf(msg, "Not enough memory to allocate a new wavelet-engine!\n");
        }
    return(E);
}

/*------------------------------------------------------------------------------------*/
long getSamplerateENGINE(ENGINE E)
{
    return E->samplerate;
}

/*------------------------------------------------------------------------------------*/
void releaseENGINE(ENGINE* E)
{
	short	i;
	if (E)
    	if (*E)
        	{
			for (i=0; i<kNumEnvelopeShapes; i++)	/* Release all envelope   */
				if ((*E)->envTables[i])				/* tables that are there. */
					free((*E)->envTables[i]);
        	if ((*E)->sineTable)
				free((*E)->sineTable);
        	releaseWAVELETS((*E)->rootWAVELET.nxt);	/* Cleanup linked list with wavelets. */
        	free(*E);                               /* Release engine-struct itself.      */
			*E = (ENGINE)NULL;						/* Clear reference. */
        	}
}
