/*
	Spectroscopic Toolkit version 1.60 by Pieter Suurmond, december 1, 2002.
    
    ST translates spectroscopic (i.e. quantumphysical) data to AIFF-file. 
	Uses the binary database-file gfall.pbin, produced by the pconv-program 
	that reads Robert Kurucz's latest data file (see subdirectory pconv).
*/

#include <stdio.h>        		/* Standard headers. */
#include <stdlib.h>
#include <math.h>

#include "ST_wavelets.h"		/* Prototype addWAVELET().  */

#include "gfall.pbin.h"        	/* Defines structure of the binary database-file,    */
                              	/* same header as included in the 'pconv'-program.   */
#include "ST_more.pbin.h"		/* To extract some more out of the database.         */
								/* But also path to the database. 					 */
#include "c00.h"		        /* All the composition-prototypes (just checking).   */

/*
    Both duration and time of the first inflection of the individual wavelet depends 
    on its' audio-frequency. First argument 'selected' in spectroscopic notation: 
    100 * Atomic number + ionisation level. For example: 100=H, 7900=Au, 4700=Ag.
    First wavelet's first inflection (up) exactly at argument start.
    Last wavelet's last inflection (down) exactly at start + duration.
    Argument duration acts as a linear time-stretch/compression parameter, it affects 
    individual wavelet-durations as well as their mutual distance in time (to the same degree).
*/
/*---------------------------------------------------------------------------------------------*/
short c02_smear(unsigned short  selected,       /* 100 * Atomic number  + ionisation level.    */
                double          strength,       /* Linear overall amplitude 3.20 is ok for Au. */
                double          start,          /* Starttime in seconds.                       */
                double          duration,       /* 'effective duration' (whatever that may mean) in seconds. */
                short           up,             /* 0 for down, 1 for up. */
                short           jj,             /* Scrambles TOP times if nonzero (with J levels). */
                ENGINE          E,
                FILE*           msg)            /* NULL, stdout or some logfile.       */
{                                               /* Just 1 element a time / time-smear. */
    FILE            *pbinDatabaseFP;
    pieterPacked    pbinBuff;
    derivedData     more;                               /* See ST.gfall.pbin.c. */
    unsigned long   linesRead = 0, linesUsed = 0, linesSkipped = 0;
    short           e = 0;
    double          audioHertz, a, dur, p, inflectStart, d, top;
    /*------------------------------------------------------------ Frequency transposition: */
    double          transpositionRatio = pow(2.0, -40.0);       /* 40 octaves down.                  */
    /*------------------------------------------------------------ Smoother frequency windowing:     */
    double          lowestFreq = 20.0;                          /* (trapeziod instead of brickwall)  */
    double          aboveLowestFreq = 1.25 * lowestFreq;        /* Rise to 1 within 1 major third.   */
    double          aboveHertz = aboveLowestFreq - lowestFreq;
    double          nyquist = 0.5*(double)getSamplerateENGINE(E);   /* Private data. */
    double          belowNyquist = nyquist / 1.2;               /* Fall to 0 within 1 minor third.   */
    double          belowHertz = nyquist - belowNyquist;
    /*------------------------------------------------------------ Open binary version of Kurucz's file. */
    double          loFreq =  9.0e99,    /* AUDIO-freq. */
                    hiFreq = -9.0e99,
                    loAmp  =  9.0e99,    /* Avalue, not audio-amplitude yet. */
                    hiAmp  = -9.0e99,
                    minJ   =  9.0e99,
                    maxJ   = -9.0e99;
    double          durA, durB;         /* Slope & offset. Lowest freq->1; highest freq->0.5. */
    
    pbinDatabaseFP = fopen(kDbase, "rb");
    if (pbinDatabaseFP)
        {
        /*------------------------------ CYCLE #1 (analysis) ------------------*/
        printf("c02_smear() analyzing database \"%s\"...\n", kDbase);
        while (fread(&pbinBuff, sizeof(pbinBuff), 1, pbinDatabaseFP))
            {            
            if (selected == pbinBuff.ELEM)
                {
                moreLineData(&pbinBuff, &more);        /* Calc some more data. */
                audioHertz = more.frequency * transpositionRatio;
                                            /* Exclude out-of-range-freqs from analysis. */
                                            /* There could slip through some freqs that are too weak! */
                if ((audioHertz > lowestFreq) && (audioHertz < nyquist))
                    {
                    if (audioHertz < loFreq) loFreq = audioHertz;
                    if (audioHertz > hiFreq) hiFreq = audioHertz;
                    if (more.Avalue < loAmp) loAmp = more.Avalue;
                    if (more.Avalue > hiAmp) hiAmp = more.Avalue;
                    
                    if (pbinBuff.LOW.J < minJ) minJ = pbinBuff.LOW.J;
                    if (pbinBuff.UPP.J < minJ) minJ = pbinBuff.UPP.J;
                    if (pbinBuff.LOW.J > maxJ) maxJ = pbinBuff.LOW.J;
                    if (pbinBuff.UPP.J > maxJ) maxJ = pbinBuff.UPP.J;
                    }
                }
            }
        /*--------------------- Print statistics for the selected element/ion: -------*/
        if (msg)
            fprintf(msg, "-------- c02_smear() ELEM %d: --------\n\
loAmp = %.8f; hiAmp = %.8f; RATIO=%.2f.\n\
loFreq = %.6f; hiFreq = %.6f; RATIO=%.2f.\n\
minJ = %.2f; maxJ = %.2f.\n",
            selected,
            loAmp,  hiAmp,  hiAmp  / loAmp,
            loFreq, hiFreq, hiFreq / loFreq,
            minJ, maxJ);
        /*---------------- Linear duration compensation function: ------------------------------*/
        /* d(loF) = 1           1 = a loF + b          a = 0.5 / (loF - hiF)                    */
        /* d(hiF) = 0.5       0.5 = a hiF + b  -     0.5 = (0.5 / (loF - hiF)) hiF + b          */
        /* d(F) = a F + b     -------------------      b = 0.5 - (0.5 hiF / (loF - hiF))        */
        /*                    0.5 = a (loF - hiF)                                               */
        d = loFreq - hiFreq;
        durA = 0.5 / d;                                 /* A=slope and B=offset. */
        durB = 0.5 - (0.5 * hiFreq / d);
        /*---------------- Linear bass-boost-filter-function (per element): --------------------*/
        /* a(loF) = 1             SAME d as with duration may be used!                          */
        /* d(hiF) = 0.5 (-6dB)  0.5 = a hiF + b  -                                              */
        /*------------------------------ CYCLE #2 ------------------*/
        rewind(pbinDatabaseFP);
        if (msg)
            {
            fprintf(msg, "1/transpositionRatio = %.1f.\n", 1.0 / transpositionRatio);
            fprintf(msg, "Effective duration = %.6f s.\n", duration);
            }
        printf("c02_smear() posting wavelets...\n");
        while (fread(&pbinBuff, sizeof(pbinBuff), 1, pbinDatabaseFP))
            {
            if (selected == pbinBuff.ELEM)
                {
                moreLineData(&pbinBuff, &more);                 /* Calc some more data. */
                /*----------------------------------------------------------------------*/
                audioHertz = more.frequency * transpositionRatio;
                if ((audioHertz > lowestFreq) && (audioHertz < nyquist))
                    {
                    /*------------------ Less brickwall-shaped filtering: --------------*/
                    if (audioHertz < aboveLowestFreq)
                        a = (audioHertz - lowestFreq) / aboveHertz; /* Rise to 1 within 1 minor third. */
                    else if (audioHertz > belowNyquist)
                        a = (nyquist - audioHertz) / belowHertz;    /* Fall to 0 within 1 major third. */
                    else
                        a = 1.0;
                    /*---------------- Frequency-dependant Compensation per element: ---------------*/
                    /* Only lowest freq of one metal (section) gets unit-duration and unit-amplit., */
                    /* all higher frequencies sound shorter.  loFreq -> 1,  hiFreq -> 0.5.          */
                    d = (durA * audioHertz) + durB;
                    /*------------------ Bass boost-filter (per element): --------------*/
                    a *= d;           /* Highest frequency -6.021 dB.                   */
                    /* ----------------- Amplitude: ------------------------------------*/
                    /* Compress amplitude-range (per section) by approx 4th-power-root. */
                    a *= strength * pow(1.90e-13 * more.Avalue, 0.26);
                    if (a >= 1.192093E-7)               /* Amplitude treshold (23 bit). */
                        {
                        /*------------------------ Panning: ---------------------------------------*/
                        if (pbinBuff.LOW.J == pbinBuff.UPP.J)
                            p = 0.0;
                        else if (pbinBuff.LOW.J > pbinBuff.UPP.J)   /* Depend upon quantum-states. */
                            p = +0.5;
                        else
                            p = -0.5;
                        if (linesUsed & 1)      /* Alternate between left and right channel.       */
                            p += 0.118;         /* Odd ones slighty to the right.     */
                        else
                            p -= 0.118;         /* Even ones  slighty to the left. */
                        /*---------------- Duration & starttime: ---------------------------*/
                        /* Only lowest frequency gets unit-duration. */
                        dur = 0.062 * d * duration;  
                        /* 0.062 defines ratio between duration and spacing of wavelets. */

                        if (up)
                            inflectStart = start + ((duration - dur) * 2.0 * (1.0 - d));
                        else /* (down) */
                            inflectStart = start + ((duration - dur) * 2.0 * (d - 0.5));
                        
                        if (jj) /* Scramble TOP times (start and end times not exact any longer!). */
                            top = inflectStart + (pbinBuff.LOW.J * pbinBuff.UPP.J * dur);
                        else
                            top = inflectStart + (0.5 * dur);
/* From low to high: */
                        if (addWAVELET(E, msg,
              /* Hz  */     audioHertz,
              /* amp */     a,
              /* pan */     p,             /* Was "pbinBuff.LOW.J * pbinBuff.UPP.J" instead of "0.5". */
              /* top */     top,            /* First inflection instead of top. */
              /* dur */     dur,
			  /* env */		kENV_Gaussian,              /* (also possible: kENV_ExpDecay with sync=0) */
			  /* syn */		1))			                /* 1 = sync with all other oscillators.       */
                            {
                            printf("Illegal wavelet!\n");
                            e = 3;
                            goto clFile;
                            }
                        linesUsed++;
                        }
                    else
                        {
                        linesSkipped++;
                        if (msg) fprintf(msg, "SKIPPED-AMP: Hz=%12.6f lggf=%5.2f.\n",
                                              audioHertz, pbinBuff.LOGGF);
                        }
                    }
                else
                    {
                    linesSkipped++;
                    if (msg) fprintf(msg, "SKIPPED-FRQ: Hz=%12.6f lggf=%5.2f.\n",
                                          audioHertz, pbinBuff.LOGGF);
                    }
                printLineEssential(&pbinBuff, msg);        /* stdout or logFP. */
                printLineDerived(&more, msg);
                }
            linesRead++;
            }
        if (msg)
            fprintf(msg, "   linesUsed = %ld.\nlinesSkipped = %ld.\n\n", linesUsed, linesSkipped);
        if (linesRead != kKuruczLinesTotal)
            {
            printf("Error reading binary database lines!\n");
            e = 2;
            }
clFile: fclose(pbinDatabaseFP);
        }
    else
        {
        printf("Cannot open binary database-file \"%s\"!\n", kDbase);
        e = 1;
        }
    return(e);
}
