/*
	Spectroscopic Toolkit version 1.40 by Pieter Suurmond, november 11, 2002.
*/

#include <stdio.h>        		/* Standard headers. */
#include <stdlib.h>
#include <math.h>

#include "gfall.pbin.h"        	/* Defines structure of the binary database-file,    */
                              	/* same header as included in the 'pconv'-program.   */
#include "ST_more.pbin.h"		/* Additional datastructure, prototypes and info.    */

/*-------------------------------------------------------------------------------------*/
void printLineEssential(pieterPacked* line, FILE* to)
{
    if (to)                                             /* Called in composition-funcs.*/
        fprintf(to, "ELEM=%6.2f; WLVAC=%11.4f; LOGGF=%7.3f; REF=%4s; \
E_LOW=%12.3f J_LOW=%5.2f %10s; E_UPP=%12.3f J_UPP=%5.2f %10s.\n",
            (double)line->ELEM / 100.0, line->WLVAC, line->LOGGF, line->REF,
            line->LOW.E, line->LOW.J, line->LOW.LABEL,
            line->UPP.E, line->UPP.J, line->UPP.LABEL);
}

/*-------------------------------------------------------------------------------------*/
void moreLineData(pieterPacked* essential, derivedData* derived)
{
    double sig, sigquad, nair;      /* No energy levels in eV, we don't need those.    */
    /*---------------------------------------------------------------------------------*/
    /* Dispersion of air / "air wavelength": B. Edlen, Metrologia V. 2, p. 71 (1966):  */
    /*---------------------------------------------------------------------------------*/
    sig = 1.0e3 / essential->WLVAC;                         /* from nm to 1/micrometer */
    sigquad = sig * sig * 1.0e8;
    nair = 1.0000834213 + 2406030.0 / (1.3e10 - sigquad) + 15997.0 / (3.89e9 - sigquad);
    if(essential->WLVAC < 200.0)                            /* nm */
        {
        derived->wlVacCorrected = essential->WLVAC;         /* WLVAC was only for vac. */
        derived->wlAirCorrected = essential->WLVAC / nair;  /* when below 200 nm!      */
        }
    else
        {
        derived->wlVacCorrected = essential->WLVAC * nair;  /* Not absolutely correct, */
        derived->wlAirCorrected = essential->WLVAC;         /* but accuracy should be  */
        }                                                   /* sufficient.             */
    /*---------------------------------------------------------------------------------*/
    /* Adapted from source file "read_pac.c" from Peter L. Smith, Claas Heise,         */
    /* Jim R. Esmond, Robert L. Kurucz (Hannover University / Harvard).                */
    /* Reference for all constants & conversions: Spectrophysics, Anne P. Thorne,      */
    /* 2nd ed., Chapman and Hall, London - New York (1988). Conversion constants for   */
    /* corrections from Astrophysical Quantities, C.W.Allen, 3rd Ed., 1976.            */
    /* A-value: A_{21} = (6.670e13 * g1/g2 / lambda/lambda) * (pow(10.0,log(gf)) / g1) */
    /*                   g1: (2*j_low + 1), g2: (2*j_upp + 1), log_gf = log (g1*f).    */
    /*---------------------------------------------------------------------------------*/
    derived->waveNumbers = essential->UPP.E - essential->LOW.E;         /* cm^-1       */
    /* Note these energy-differences are sometimes NEGATIVE!!(??)                      */
    /* (Except from sign, corresponding wavelengths seem however to be ok.)            */
    if (derived->waveNumbers < 0)
        {
        derived->waveNumbers = (double)-1.0 * derived->waveNumbers;
   /*     printf("(\"moreLineData()\" negated frequency)\n");   */
        }
    derived->photonEnergy =  1.2398544e-4 * derived->waveNumbers;       /* cm^-1 -> eV */
    derived->frequency = 2.417965e14 * derived->photonEnergy;           /* s^-1        */
    derived->Avalue = 6.670e13 / (2.0 * essential->UPP.J + 1.0) / 
                      derived->wlVacCorrected / derived->wlVacCorrected *
                      pow(10.0, essential->LOGGF);
}

/*-------------------------------------------------------------------------------------*/
void printLineDerived(derivedData* line, FILE* to)
{
    if (to)
        fprintf(to, "           vacCorr=%11.4f; airCorr=%11.4f; frq=%.10e; A=%.4e.\n",
            line->wlVacCorrected, line->wlAirCorrected, line->frequency, line->Avalue);
}
