/*******************************************************************************
   pconv.c - Reads Kurucz's latest data file, Pieter Suurmond, nov 11, 2002.
   To run, it needs "gfall.dat" so first gunzip the gzipped "gfall.dat.gz"!
   Linelists with names of the form GF* have the following 160 column format:
1                                                                             80
+++++++++++^^^^^^^++++++^^^^^^^^^^^^+++++^++++++++++^^^^^^^^^^^^+++++^++++++++++
   800.7110  0.116 27.00   45924.947  3.5 (3F)5s e2F   33439.661  4.5 (3F)4p y2G
    wl(nm)  log gf elem      E(cm-1)   J   label        E'(cm-1)   J'   label'  
                   code [char*28 level descriptor  ][char*28 level descriptor  ]
81                                                                           160
^^^^^^++++++^^^^^^++++^^++^^^++++++^^^++++++^^^^^+++++^+^+^+^+++^^^^^+++++^^^^^^
  8.19 -5.38 -7.59K88  0 0 59-2.584 59 0.000  104  -77F6 -5 0    1140 1165     0
  log   log   log ref NLTE iso log iso  log     hyper  F F'     glande       iso
 Gamma Gamma Gamma   level hyper f iso frac   shift(mK)     ^       glande'shift
  rad  stark  vdW    numbers                    E    E'     ^abc  (x1000)   (mA)
                                                         I*1^char*3
 1 wavelength (nm)  air above 200 nm   F11.4               codes
 2 log gf  F7.3
 3 element code = element number + charge/100.  F6.2
 4 first energy level in cm-1   F12.3
 5 J for first level   F5.1
   blank for legibility   1X
 6 label field for first level   A10
 7 second energy level in cm-1   F12.3
        (negative energies are predicted or extrapolated}  
 8 J for second level   F5.1
   blank for legibility   1X
 9 label field for second level   A10
10 log of radiative damping constant, Gamma Rad  F6.2 or F6.3
11 log of stark damping constant/electron number. Gamma Stark  F6.2 or F6.3
12 log of van der Waals damping constant/neutral hydrogen number, 
       Gamma van der Waals   F6.2 or F6.3
13 reference that can be expanded in subdirectory LINES   A4  
14 non-LTE level index for first level   I2
15 non-LTE level index for second level   I2
16 isotope number   I3
17 hyperfine component log fractional strength  F6.3
18 isotope number  (for diatomics there are two and no hyperfine)   I3
19 log isotopic abundance fraction   F6.3
20 hyperfine shift for first level in mK to be added to E  I5
21 hyperfine shift for second level in mK to be added to E'  I5
   the symbol "F" for legibilty   1X
22 hyperfine F for the first level    I1
23 note on character of hyperfine data for first level: z none, ? guessed  A1
   the symbol "-" for legibility    1X
24 hyperfine F' for the second level  I1
25 note on character of hyperfine data for second level: z none, ? guessed  A1
26 1-digit code, sometimes for line strength classes   I1
27 3-character code such as AUT for autoionizing    A3  
28 lande g for first level times 1000   I5
29 lande g for second level times 1000   I5
30 isotope shift of wavelength in mA 

FORMAT(F11.4,F7.3,F6.2,F12.3,F5.2,1X,A10,F12.3,F5.2,1X,A10,
       3F6.2,A4,2I2,I3,F6.3,I3,F6.3,2I5,1X,A1,A1,1X,A1,A1,i1,A3.2I5,I6) 
********************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "gfall.pbin.h"   /* Defines format of binary-file.   */

/*------------------------------------------------------*/
int processLine(char* textBuff, pieterPacked* PBIN)
{
    char    rem;
    short   n;
    int     err = 0;
    /*--------------------------------------- WLVAC: 11 */
    rem = textBuff[11];
    textBuff[11] = 0;
    PBIN->WLVAC = strtod(&textBuff[0], (char**)NULL);
    textBuff[11] = rem;
    /*--------------------------------------- LOGGF: 7  */
    rem = textBuff[18];
    textBuff[18] = 0;
    PBIN->LOGGF = strtod(&textBuff[11], (char**)NULL);
    textBuff[18] = rem;
    /*--------------------------------------- ELEM: 6  (to int) */
    rem = textBuff[24];
    textBuff[24] = 0;
    PBIN->ELEM = (short)(0.5 + (100.0 * strtod(&textBuff[18], (char**)NULL)));
    textBuff[24] = rem;
    /*--------------------------------------- E_LOW: 12 */
    rem = textBuff[36];
    textBuff[36] = 0;
    PBIN->LOW.E = strtod(&textBuff[24], (char**)NULL);
    textBuff[36] = rem;
    /*--------------------------------------- J_LOW: 5 */
    rem = textBuff[41];
    textBuff[41] = 0;
    PBIN->LOW.J = (float)strtod(&textBuff[36], (char**)NULL);
    textBuff[41] = rem;
    /*--------------------------------------- blank: 1 (check!) */
    if (textBuff[41] != ' ')
        err |= 1;
    /*--------------------------------------- LABEL_LOW: 10 */
    n = 52;
    while ((textBuff[n-1] == ' ') && (n > 42))
        n--;
    rem = textBuff[n];
    textBuff[n] = 0;
    strcpy(PBIN->LOW.LABEL, &textBuff[42]);
    textBuff[n] = rem;
    /*--------------------------------------- E_UPP: 12 */
    rem = textBuff[64];
    textBuff[64] = 0;
    PBIN->UPP.E = strtod(&textBuff[52], (char**)NULL);
    textBuff[64] = rem;
    /*--------------------------------------- J_UPP: 5 */
    rem = textBuff[69];
    textBuff[69] = 0;
    PBIN->UPP.J = (float)strtod(&textBuff[64], (char**)NULL);
    textBuff[69] = rem;
    /*--------------------------------------- blank: 2 (check!) */
    if (textBuff[69] != ' ')
        err |= 2;
    /*--------------------------------------- LABEL_UPP: 10 */
    n = 80;
    while ((textBuff[n-1] == ' ') && (n > 70))
        n--;
    rem = textBuff[n];
    textBuff[n] = 0;
    strcpy(PBIN->UPP.LABEL, &textBuff[70]);
    textBuff[n] = rem;
    /*--------------------------------------- REF: 4 */
    n = 102;
    while ((textBuff[n-1] == ' ') && (n > 98))
        n--;
    rem = textBuff[n];
    textBuff[n] = 0;
    strcpy(PBIN->REF, &textBuff[98]);
    textBuff[n] = rem;
    /*---------------------------------------*/
    return(err);
}

/*------------------------------*/
#define kKuruczLineLength 160
#define kKuruczLinesTotal 564942L

/*------------------------------*/
int main(int argc, char *argv[])
{
    const char      *textFilename = "gfall.dat",        /* Pathname of INPUT-FILE.  */
                    *pbinFilename = "gfall.pbin";       /* Pathname of OUTPUT-FILE. */
    FILE            *textDatabaseFP,
                    *pbinDatabaseFP;
    char            lineBuff[kKuruczLineLength + 4];    /* +4 for LF, CR and Z.     */
    pieterPacked    pbinBuff;
    unsigned long   linesParsed;
    int             e;

    textDatabaseFP = fopen(textFilename, "rt");         /* OPEN FILE (WAS JUST "r", now "rt"). */
    if (textDatabaseFP)
        {
        printf("Reading textfile \"%s\".\n", textFilename);
        pbinDatabaseFP = fopen(pbinFilename, "wb");                      /* CREATE FILE. */
        if (pbinDatabaseFP)
            {
            printf("Writing binary file '%s', %d bytes per record.\n",
                pbinFilename, (int)sizeof(pbinBuff));
            linesParsed = 0L;
            while (fgets(lineBuff, kKuruczLineLength + 2, textDatabaseFP))  /* +1 +1 */
                {
                e = processLine(lineBuff, &pbinBuff);   /* READ FROM FILE. */
                if (e)
                    printf("Error in processLine (ret %d), at line #%ld!\n",
                        e, 1L+linesParsed);
                if ((pbinBuff.ELEM < 100) || (pbinBuff.ELEM > 9201))
                    printf("Error: element out of range, line #%ld!\n", 1L+linesParsed);
                if ((pbinBuff.WLVAC < 2.50) || (pbinBuff.WLVAC > 754956.3) ||
                    (pbinBuff.LOGGF > 3.75) || (pbinBuff.LOGGF < -18.0))
                    printf("#%6ld: ELEM=%6.2f; WLVAC=%11.4f; LOGGF=%7.3f; REF=%s;\n\
         E_LOW=%12.3f J_LOW=%5.2f %s;\n\
         E_UPP=%12.3f J_UPP=%5.2f %s.\n",
                    1L+linesParsed,
                    (double)pbinBuff.ELEM / 100.0, pbinBuff.WLVAC, pbinBuff.LOGGF, pbinBuff.REF,
                    pbinBuff.LOW.E, pbinBuff.LOW.J, pbinBuff.LOW.LABEL,
                    pbinBuff.UPP.E, pbinBuff.UPP.J, pbinBuff.UPP.LABEL);
                fwrite(&pbinBuff, sizeof(pbinBuff), 1, pbinDatabaseFP);     /* WRITE TO FILE. */
                linesParsed++;
                }
            if (linesParsed != kKuruczLinesTotal)
                printf("Too few or too many lines in database file: linesParsed = %ld.\n",
                    linesParsed);
            else
                printf("Parsed right amount of lines (%ld), binary file should be %ld bytes.\n",
                    linesParsed, linesParsed * sizeof(pbinBuff));
            fclose(pbinDatabaseFP);
            }
        else
            printf("Cannot create file \"%s\".\n", pbinFilename);
        fclose(textDatabaseFP);
        }
    else
        printf("Cannot open file \"%s\".\n", textFilename);
    return 0;
}
